from dataclasses import asdict

from sqlalchemy import select

from expanse.database.session import Session
from expanse.http.helpers import json
from expanse.http.response import Response
from expanse.routing.router import Router
from expanse.testing.client import TestClient
from tests.synchronous.integration.database.models import User


def show(session: Session, user_email: str) -> Response:
    user = session.scalars(select(User).where(User.email == user_email).limit(1)).one()

    return json(asdict(user))


def test_no_result_found_errors_return_404_errors(router: Router, client: TestClient):
    router.get("/users/{user_email}", show)

    response = client.get("/users/john@doe.com", headers={"Accept": "application/json"})

    assert response.status_code == 200

    response = client.get("/users/jane@doe.com", headers={"Accept": "application/json"})

    assert response.status_code == 404
    assert response.json() == {
        "exception": "HTTPException",
        "message": "No row was found when one was required",
    }



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Protocol


if TYPE_CHECKING:
    from expanse.testing.command_tester import TestingCommand


class TestingCommandHelper(Protocol):
    def __call__(self, name: str, parameters: str | None = None) -> TestingCommand: ...



from expanse.asynchronous.console.commands.command import Command


class FooBarCommand(Command):
    name = "foo bar"

    async def handle(self) -> int:
        self.info("Foo Bar")

        return 0



from expanse.asynchronous.http.response import Response
from expanse.asynchronous.http.response_adapter import ResponseAdapter
from expanse.asynchronous.support.service_provider import ServiceProvider
from expanse.asynchronous.view.view import View
from expanse.asynchronous.view.view_factory import ViewFactory
from expanse.asynchronous.view.view_finder import ViewFinder
from expanse.common.configuration.config import Config


class ViewServiceProvider(ServiceProvider):
    async def register(self) -> None:
        await self._register_factory()
        await self._register_view_finder()

    async def boot(self) -> None:
        await self._container.on_resolved(
            ResponseAdapter, self._register_response_adapters
        )

    async def _register_factory(self) -> None:
        self._container.singleton(ViewFactory, self._create_factory)
        self._container.alias(ViewFactory, "view")

    async def _create_factory(self, finder: ViewFinder, config: Config) -> ViewFactory:
        return ViewFactory(finder, debug=config["app"].get("debug", False))

    async def _register_view_finder(self) -> None:
        async def _create_view_finder(config: Config) -> ViewFinder:
            return ViewFinder(config["view"]["paths"])

        self._container.singleton(ViewFinder, _create_view_finder)
        self._container.alias(ViewFinder, "view:finder")

    async def _register_response_adapters(self, adapter: ResponseAdapter) -> None:
        async def adapt_view(raw_response: View, factory: ViewFactory) -> Response:
            return await factory.render(raw_response)

        adapter.register_adapter(View, adapt_view)



from abc import ABC
from typing import Generic
from typing import Self
from typing import TypeVar

from expanse.common.support._utils import class_to_name


T = TypeVar("T")


class ServiceProvidersList(Generic[T], ABC):
    def __init__(self, providers: list[type[T] | str]) -> None:
        self._providers: list[str] = [class_to_name(provider) for provider in providers]

    def merge(self, providers: list[type[T] | str]) -> Self:
        return self.__class__([*self._providers, *providers])

    def without(self, providers: list[type[T] | str]) -> Self:
        provider_names: set[str] = {class_to_name(provider) for provider in providers}

        new_providers: list[str | type] = []
        for current_provider in self._providers:
            if current_provider in provider_names:
                continue

            new_providers.append(current_provider)

        return self.__class__(new_providers)

    def to_list(self) -> list[str]:
        return self._providers.copy()

    @classmethod
    def default(cls) -> Self:
        return cls([])



from alembic import context

from expanse.common.database.orm.model import Model
from expanse.database.database_manager import DatabaseManager


config = context.config
db: DatabaseManager = context.config.attributes["db"]

target_metadata = Model.metadata


def run_migrations_offline() -> None:
    engine = db.configure_engine()

    context.configure(
        url=engine.url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    with db.connection() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from expanse.http.request import Request


def test_url() -> None:
    request = Request.create("http://example.com:1234/foo/bar?foo=bar&bar=baz")

    url = request.url

    assert url.scheme == "http"
    assert url.netloc == "example.com:1234"
    assert url.path == "/foo/bar"
    assert url.query == "foo=bar&bar=baz"
    assert url.fragment == ""
    assert url.username is None
    assert url.password is None
    assert url.hostname == "example.com"
    assert url.port == 1234
    assert url.full == "http://example.com:1234/foo/bar?foo=bar&bar=baz"



from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING
from typing import Annotated
from typing import Any
from typing import get_args
from typing import get_origin

from pydantic import BaseModel

from expanse.common.http.form import Form
from expanse.common.http.json import JSON
from expanse.common.http.query import Query
from expanse.common.http.url_path import URLPath
from expanse.common.routing.exceptions import RouteNotFound
from expanse.common.routing.route_matcher import RouteMatcher
from expanse.container.container import Container
from expanse.http.request import Request
from expanse.http.response import Response
from expanse.http.response_adapter import ResponseAdapter
from expanse.routing.pipeline import Pipeline
from expanse.routing.route import Route
from expanse.routing.route_collection import RouteCollection
from expanse.routing.route_group import RouteGroup
from expanse.types.http.middleware import RequestHandler
from expanse.types.routing import Endpoint


if TYPE_CHECKING:
    from collections.abc import Callable


class Router:
    def __init__(self, container: Container) -> None:
        self._container: Container = container
        self._routes: RouteCollection = RouteCollection()

    @property
    def routes(self) -> RouteCollection:
        return self._routes

    def add_route(self, route: Route) -> None:
        self._routes.add(route)

    def add_routes(self, routes: list[Route]) -> None:
        for route in routes:
            self.add_route(route)

    def add_group(self, group: RouteGroup) -> None:
        for route in group.routes:
            self.add_route(route)

    def add_groups(self, groups: list[RouteGroup]) -> None:
        for group in groups:
            self.add_group(group)

    def get(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.get(path, endpoint, name=name)
        self.add_route(route)

        return route

    def post(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.post(path, endpoint, name=name)
        self.add_route(route)

        return route

    def put(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.put(path, endpoint, name=name)
        self.add_route(route)

        return route

    def patch(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.patch(path, endpoint, name=name)
        self.add_route(route)

        return route

    def delete(
        self, path: str, endpoint: Endpoint, *, name: str | None = None
    ) -> Route:
        route = Route.delete(path, endpoint, name=name)
        self.add_route(route)

        return route

    def head(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.head(path, endpoint, name=name)
        self.add_route(route)

        return route

    def options(
        self, path: str, endpoint: Endpoint, *, name: str | None = None
    ) -> Route:
        route = Route.options(path, endpoint, name=name)
        self.add_route(route)

        return route

    @contextmanager
    def group(
        self,
        name: str | None = None,
        prefix: str | None = None,
    ) -> Generator[RouteGroup, None, None]:
        group = RouteGroup(name=name, prefix=prefix)

        yield group

        self.add_group(group)

    def route(self, name: str, parameters: dict[str, Any] | None = None) -> URLPath:
        parameters = parameters or {}

        for route in self._routes:
            if route.name == name:
                matcher = self._container.get(RouteMatcher)

                return matcher.url(route.path, **parameters)

        raise RouteNotFound(f"Route [{name}] is not defined")

    def url(self, path: str, parameters: dict[str, Any] | None = None) -> URLPath:
        matcher = self._container.get(RouteMatcher)

        parameters = parameters or {}

        return matcher.url(path, **parameters)

    def handle(self, container: Container, request: Request) -> Response:
        route = self._routes.match(request)

        handler: RequestHandler
        pipes: list[Callable[[Request, RequestHandler], Response]] = []
        if route is None:
            # No matching route was found, so we use the default handler
            # to handle the request.
            handler = self._default_handler(container)
        else:
            # Set the route to the request
            request.set_route(route)

            handler = self._route_handler(route, container)
            pipes = [
                container.get(middleware).handle
                for middleware in route.get_middleware()
            ]

        return Pipeline(container).use(pipes).send(request).to(handler)

    def _route_handler(self, route: Route, container: Container) -> RequestHandler:
        def handler(request: Request) -> Response:
            arguments = {}

            for name, parameter in route.signature.parameters.items():
                if name in route.param_names:
                    arguments[name] = request.path_params[name]
                elif isinstance(parameter.annotation, type) and issubclass(
                    parameter.annotation, Form
                ):
                    arguments[name] = parameter.annotation(request.form)
                elif origin := get_origin(
                    parameter.annotation
                ) is Annotated and issubclass(
                    (origin_args := get_args(parameter.annotation))[0], BaseModel
                ):
                    validation_model: type[BaseModel] = origin_args[0]
                    data_type: type[JSON] | type[Query] | JSON | Query = origin_args[1]

                    if isinstance(data_type, JSON) or issubclass(data_type, JSON):
                        arguments[name] = validation_model.model_validate(request.json)
                    elif isinstance(data_type, Query) or issubclass(data_type, Query):
                        arguments[name] = validation_model.model_validate(
                            request.query_params
                        )

            raw_response = container.call(route.endpoint, **arguments)

            # Do not go through the response adapter if the response is already a Response instance
            if isinstance(raw_response, Response):
                return raw_response

            declared_response_type = route.signature.return_annotation

            return container.get(ResponseAdapter).adapt(
                raw_response, declared_response_type=declared_response_type
            )

        return handler

    def _default_handler(self, container: Container) -> RequestHandler:
        def handler(request: Request) -> Response:
            from expanse.routing.responder import Responder

            container.get(Responder).abort(404)

        return handler



from typing import TYPE_CHECKING

from expanse.asynchronous.console.commands.command import Command
from expanse.asynchronous.routing.router import Router


if TYPE_CHECKING:
    from expanse.asynchronous.routing.route import Route

METHODS_COLORS: dict[str, str] = {
    "GET": "fg=blue",
    "HEAD": "fg=default;options=dark",
    "POST": "fg=yellow",
    "PUT": "fg=yellow",
    "PATCH": "fg=yellow",
    "DELETE": "fg=red",
    "OPTIONS": "fg=default;options=dark",
}


class RoutesListCommand(Command):
    name: str = "routes list"

    async def handle(self, router: Router) -> int:
        routes: list[Route] = []

        max_methods_length = 0

        for route in router.routes:
            max_methods_length = max(
                max_methods_length,
                sum(len(method) for method in route.methods) + len(route.methods) - 1,
            )
            routes.append(route)

        routes.sort(key=lambda route: route.path)

        self.line("")

        for route in routes:
            methods_length = (
                sum(len(method) for method in route.methods) + len(route.methods) - 1
            )
            methods = "|".join(
                f"<{METHODS_COLORS.get(method.upper(), 'fg=default')}>{method}</>"
                for method in route.methods
            )
            line = [
                "  ",
                methods,
                " " * (max_methods_length - methods_length + 2),
                route.path,
            ]

            if route.name:
                line.append(f" <options=dark>({route.name})</>")

            self.line("".join(line))

            if self._io.is_verbose():
                padding = " " * (methods_length + max_methods_length - methods_length)
                endpoint = route.endpoint.__module__ + "." + route.endpoint.__qualname__
                self.line(f"  {padding}  <options=dark>{endpoint}</>")

        return 0



from __future__ import annotations

import pytest


@pytest.fixture()
def scope():
    return {
        "type": "http",
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "http_version": "1.1",
        "server": ("127.0.0.1", 8000),
        "client": ("127.0.0.1", 51715),
        "scheme": "http",
        "root_path": "",
        "headers": [
            (b"host", b"127.0.0.1:8000"),
            (b"accept-encoding", b"gzip, deflate"),
            (b"accept", b"*/*"),
            (b"connection", b"keep-alive"),
            (b"user-agent", b"Foo/3.2.2"),
        ],
        "state": {},
        "method": "GET",
        "path": "/",
        "raw_path": b"/",
        "query_string": b"",
    }



from typing import Any
from typing import TypeVar
from typing import overload

from sqlalchemy import CursorResult
from sqlalchemy import Executable
from sqlalchemy import Result
from sqlalchemy import ScalarResult
from sqlalchemy import UpdateBase
from sqlalchemy import text
from sqlalchemy import util
from sqlalchemy.engine.interfaces import _CoreAnyExecuteParams
from sqlalchemy.ext.asyncio import AsyncSession as BaseAsyncSession
from sqlalchemy.orm._typing import OrmExecuteOptionsParameter
from sqlalchemy.orm.session import _BindArguments
from sqlalchemy.sql.selectable import TypedReturnsRows


_T = TypeVar("_T", bound=Any)


class Session(BaseAsyncSession):
    @overload  # type: ignore[override]
    async def execute(
        self,
        statement: UpdateBase,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> CursorResult[Any]: ...

    @overload
    async def execute(
        self,
        statement: Executable,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> Result[Any]: ...

    @overload
    async def execute(
        self,
        statement: str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: dict[str, Any] | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> Result[Any]: ...

    async def execute(
        self,
        statement: UpdateBase | Executable | str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> Result[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return await super().execute(
            statement,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            _parent_execute_state=_parent_execute_state,
            _add_event=_add_event,
        )

    @overload
    async def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> _T | None: ...

    @overload
    async def scalar(
        self,
        statement: Executable,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> Any: ...

    @overload
    async def scalar(
        self,
        statement: str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> Any: ...

    async def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]] | Executable | str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> Any:
        if isinstance(statement, str):
            statement = text(statement)

        return await super().scalar(
            statement,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            **kw,
        )

    @overload
    async def scalars(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> ScalarResult[_T]: ...

    @overload
    async def scalars(
        self,
        statement: Executable,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> ScalarResult[Any]: ...

    @overload
    async def scalars(
        self,
        statement: str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: dict[str, Any] | None = None,
        **kw: Any,
    ) -> ScalarResult[Any]: ...

    async def scalars(
        self,
        statement: Executable | str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: dict[str, Any] | None = None,
        **kw: Any,
    ) -> ScalarResult[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return await super().scalars(
            statement,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            **kw,
        )



from abc import ABC
from abc import abstractmethod
from collections.abc import Callable
from collections.abc import Generator
from contextlib import contextmanager
from typing import Self
from typing import TypeVar

from cleo.io.outputs.output import Output

from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response


_TError = TypeVar("_TError", bound=Exception)


class ExceptionHandler(ABC):
    @abstractmethod
    async def report(self, e: Exception) -> None:
        """
        Report or log an exception.

        :param e: The exception to report or log

        :raises Exception
        """

    @abstractmethod
    async def should_report(self, e: Exception) -> bool:
        """
        Determine whether the exception should be reported or not.

        :param e: The exception to verify
        """
        ...

    @abstractmethod
    def ignore(self, exception_class: type[Exception]) -> Self:
        """
        Ignore a specific exception class when reporting exceptions.

        :param exception_class: The exception class to ignore.
        """
        ...

    @abstractmethod
    async def render(self, request: Request, e: Exception) -> Response:
        """
        Render the exception into an HTTP response.

        :param request: The request during which the exception was raised
        :param e: The exception to render
        """
        ...

    @abstractmethod
    async def render_for_console(self, output: Output, e: Exception) -> None:
        """
        Render the exception to the console output.

        :param output: the output to render the exception to
        :param e: The exception to render
        """
        ...

    @abstractmethod
    @contextmanager
    def raise_unhandled_exceptions(
        self, raise_exceptions: bool = True
    ) -> Generator[None, None, None]:
        """
        Temporarily enable/disable raising unhandled exceptions.
        This is mainly useful for debugging purposes.

        :param raise_exceptions: Whether unhandled exceptions should be raised.
        """
        ...

    @abstractmethod
    def prepare_using(
        self, exception_class: type[_TError], preparer: Callable[[_TError], Exception]
    ) -> None:
        """
        Register a preparer for a specific exception class.

        :param exception_class: The exception class to prepare, i.e. convert to another exception.
        :param preparer: The preparer function to use to convert the given exception type.
        """
        ...



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import TypeVar

from cleo.exceptions import CleoError
from cleo.io.inputs.string_input import StringInput
from cleo.io.null_io import NullIO

from expanse.common.console.command import Command as BaseCommand


if TYPE_CHECKING:
    from cleo.io.io import IO

    from expanse.console.console import Console  # noqa: TID
    from expanse.container.container import Container

TApplication = TypeVar("TApplication", bound="Console")


class Command(BaseCommand[TApplication], ABC):
    def __init__(self) -> None:
        super().__init__()

        self._container: Container | None = None

    @abstractmethod
    def handle(self, *args, **kwargs) -> int | None: ...

    def run(self, io: IO) -> int:
        self.merge_console_definition()

        try:
            io.input.bind(self.definition)
        except CleoError:
            if not self._ignore_validation_errors:
                raise

        if io.input.has_argument("command") and io.input.argument("command") is None:
            io.input.set_argument("command", self.name)

        io.input.validate()

        status_code = self.execute(io)

        if status_code is None:
            status_code = 0

        return status_code

    def execute(self, io: IO) -> int | None:
        self._io = io

        try:
            if not self._container:
                return self.handle()

            return self._container.call(self.handle)
        except KeyboardInterrupt:
            return 1

    def set_container(self, container: Container) -> None:
        self._container = container

    def call(self, name: str, args: str | None = None) -> int:
        """
        Call another command.
        """
        if args is None:
            args = ""

        input = StringInput(args)
        assert self._console is not None
        command = self._console.get(name)

        return self._console._run_command(command, self._io.with_input(input))

    def call_silent(self, name: str, args: str | None = None) -> int:
        """
        Call another command silently.
        """
        if args is None:
            args = ""

        input = StringInput(args)
        assert self._console is not None
        command = self._console.get(name)

        return self._console._run_command(command, NullIO(input))



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Generator

    from expanse.container.container import Container


_container: ContextVar[Container | None] = ContextVar("container", default=None)


def _get_container() -> Container:
    container = _container.get()

    if container is None:
        raise RuntimeError("Container not set.")

    return container


def _set_container(container: Container | None) -> None:
    _container.set(container)


@contextmanager
def _use_container(container: Container) -> Generator[Container, None, None]:
    _set_container(container)

    yield container

    _set_container(None)



from __future__ import annotations

import traceback

from pathlib import Path
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Self

from expanse.asynchronous.container.container import Container
from expanse.asynchronous.contracts.debug.exception_handler import (
    ExceptionHandler as ExceptionHandlerContract,
)
from expanse.asynchronous.core.application_builder import ApplicationBuilder
from expanse.asynchronous.core.bootstrap.boot_providers import BootProviders
from expanse.asynchronous.core.bootstrap.load_configuration import LoadConfiguration
from expanse.asynchronous.core.bootstrap.load_environment_variables import (
    LoadEnvironmentVariables,
)
from expanse.asynchronous.core.bootstrap.register_providers import RegisterProviders
from expanse.asynchronous.exceptions.handler import ExceptionHandler
from expanse.common.configuration.config import Config
from expanse.common.core.application import Application as BaseApplication
from expanse.common.support._utils import string_to_class


if TYPE_CHECKING:
    from collections.abc import Awaitable
    from collections.abc import Callable

    from cleo.io.inputs.input import Input

    from expanse.asynchronous.core.bootstrap.bootstrapper import Bootstrapper
    from expanse.asynchronous.support.service_provider import ServiceProvider
    from expanse.asynchronous.types import Receive
    from expanse.asynchronous.types import Scope
    from expanse.asynchronous.types import Send


class Application(BaseApplication):
    _bootstrappers: ClassVar[list[type[Bootstrapper]]] = [
        LoadEnvironmentVariables,
        LoadConfiguration,
        RegisterProviders,
        BootProviders,
    ]

    def __init__(
        self, base_path: Path | None = None, container: Container | None = None
    ) -> None:
        BaseApplication.__init__(
            self,
            base_path
            or Path(traceback.extract_stack(limit=2)[0].filename).parent.parent,
        )

        if container is None:
            container = Container()

        self._container = container

        self._service_providers: list[ServiceProvider] = []
        self._default_bootstrappers: list[type[Bootstrapper]] = (
            self.__class__._bootstrappers.copy()
        )
        self._bootstrapping_callbacks: list[Callable[[Container], Awaitable[None]]] = []

        self._bind_paths()
        self._register_base_bindings()

    @property
    def container(self) -> Container:
        return self._container

    @classmethod
    def configure(cls, base_path: Path | None = None) -> ApplicationBuilder:
        base_path = (
            base_path
            or Path(traceback.extract_stack(limit=2)[0].filename).parent.parent
        )

        return ApplicationBuilder(base_path).with_kernels().with_commands()

    def set_config(self, config: Config) -> None:
        super().set_config(config)
        self._container.instance(Config, config)

    async def boot(self) -> None:
        """
        Boot the application service providers.
        """
        if self.is_booted():
            return

        for service_provider in self._service_providers:
            await self._boot_provider(service_provider)

        self._booted = True

    def set_base_path(self, base_path: Path) -> Self:
        self._base_path = base_path

        self._bind_paths()

        return self

    async def bootstrap(self) -> Self:
        return await self.bootstrap_with(self._default_bootstrappers)

    async def bootstrap_with(self, bootstrappers: list[type[Bootstrapper]]) -> Self:
        if self._has_been_bootstrapped:
            return self

        await self._register_base_service_providers()

        for bootstrapper_class in bootstrappers:
            bootstrapper: Bootstrapper = await self._container.get(bootstrapper_class)
            await bootstrapper.bootstrap(self)

        for callback in self._bootstrapping_callbacks:
            await callback(self._container)

        await self.boot()

        self._has_been_bootstrapped = True

        return self

    def bootstrapping(self, *callback: Callable[[Container], Awaitable[None]]) -> Self:
        self._bootstrapping_callbacks.extend(callback)

        return self

    async def register_configured_providers(self) -> None:
        providers = (await self._container.get(Config)).get("app.providers", [])

        for provider_class in providers:
            if isinstance(provider_class, str):
                provider_class = string_to_class(provider_class)

            provider = provider_class(self._container)

            await self.register(provider)

    async def register(
        self, provider: ServiceProvider, force: bool = False
    ) -> ServiceProvider:
        self._service_providers.append(provider)

        await provider.register()

        return provider

    async def handle_command(self, input: Input) -> int:
        from expanse.asynchronous.core.console.gateway import Gateway

        kernel = await self._container.get(Gateway)

        return await kernel.handle(input)

    def _bind_paths(self) -> None:
        assert self._base_path is not None

        self._container.instance("path", self._base_path)
        self._container.instance("path:config", self.config_path)

    async def _boot_provider(self, provider: ServiceProvider) -> None:
        if hasattr(provider, "boot"):
            await self._container.call(provider.boot)

    def _register_base_bindings(self) -> None:
        from expanse.asynchronous.core.http.gateway import Gateway

        self._container.instance(self.__class__, self)
        self._container.instance(BaseApplication, self)
        self._container.alias(self.__class__, "app")
        self._container.instance(Container, self._container)
        self._config = Config({})
        self._container.instance(Config, self._config)
        self._container.alias(Config, "config")

        self._container.singleton(Gateway)

        self._container.singleton(ExceptionHandlerContract, ExceptionHandler)

    async def _register_base_service_providers(self) -> None:
        from expanse.asynchronous.http.http_service_provider import HTTPServiceProvider
        from expanse.asynchronous.routing.routing_service_provider import (
            RoutingServiceProvider,
        )

        await self.register(HTTPServiceProvider(self._container))
        await self.register(RoutingServiceProvider(self._container))

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "lifespan":
            while True:
                message = await receive()
                if message["type"] == "lifespan.startup":
                    await self.bootstrap()
                    await send({"type": "lifespan.startup.complete"})
                elif message["type"] == "lifespan.shutdown":
                    await self._container.terminate()
                    await send({"type": "lifespan.shutdown.complete"})
                    return

        from expanse.asynchronous.core.http.gateway import Gateway

        gateway = await self._container.get(Gateway)

        await gateway(scope, receive, send)



from expanse.http.response import Response
from expanse.routing.responder import Responder
from expanse.routing.router import Router
from expanse.testing.client import TestClient


def destination() -> str:
    return "Arrived at destination"


def destination_with_parameters(name: str) -> str:
    return f"Arrived at {name} destination"


def redirection_to_url(responder: Responder) -> Response:
    return responder.redirect().to("/destination")


def redirection_to_named_route(responder: Responder) -> Response:
    return responder.redirect().to_route("destination")


def redirection_to_named_route_with_parameters(responder: Responder) -> Response:
    return responder.redirect().to_route("destination", {"name": "foo"})


def redirection_to_named_route_with_parameters_and_query_params(
    responder: Responder,
) -> Response:
    return responder.redirect().to_route(
        "destination", {"name": "foo", "foo": "bar", "bar": 42}
    )


def test_redirect_to_url(router: Router, client: TestClient) -> None:
    router.get("/destination", destination)
    router.get("/redirection", redirection_to_url)

    response = client.get("/redirection", follow_redirects=False)

    assert response.status_code == 302
    assert response.headers["Location"] == "/destination"

    response = client.get("/redirection", follow_redirects=True)

    assert response.status_code == 200
    assert response.text == '"Arrived at destination"'


def test_redirect_to_named_route(router: Router, client: TestClient) -> None:
    router.get("/destination", destination, name="destination")
    router.get("/redirection", redirection_to_named_route)

    response = client.get("/redirection", follow_redirects=False)

    assert response.status_code == 302
    assert response.headers["Location"] == "/destination"

    response = client.get("/redirection", follow_redirects=True)

    assert response.status_code == 200
    assert response.text == '"Arrived at destination"'


def test_redirect_to_named_route_with_parameters(
    router: Router, client: TestClient
) -> None:
    router.get("/destination/{name}", destination_with_parameters, name="destination")
    router.get("/redirection", redirection_to_named_route_with_parameters)

    response = client.get("/redirection", follow_redirects=False)

    assert response.status_code == 302
    assert response.headers["Location"] == "/destination/foo"

    response = client.get("/redirection", follow_redirects=True)

    assert response.status_code == 200
    assert response.text == '"Arrived at foo destination"'


def test_redirect_to_named_route_with_parameters_and_query_params(
    router: Router, client: TestClient
) -> None:
    router.get("/destination/{name}", destination_with_parameters, name="destination")
    router.get(
        "/redirection", redirection_to_named_route_with_parameters_and_query_params
    )

    response = client.get("/redirection", follow_redirects=False)

    assert response.status_code == 302
    assert response.headers["Location"] == "/destination/foo?foo=bar&bar=42"

    response = client.get("/redirection", follow_redirects=True)

    assert response.status_code == 200
    assert response.text == '"Arrived at foo destination"'



from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column

from tests.asynchronous.integration.database.console.commands.fixtures.models.model import (
    Model,
)
from tests.asynchronous.integration.database.console.commands.fixtures.models.model import (
    primary_key,
)


class User(Model):
    __tablename__: str = "users"

    id: Mapped[primary_key] = mapped_column(init=False)
    first_name: Mapped[str | None] = mapped_column(default=None)
    last_name: Mapped[str | None] = mapped_column(default=None)
    email: Mapped[str | None] = mapped_column(default=None)



from typing import Any
from typing import Self
from typing import TypeVar
from typing import overload

from sqlalchemy import Connection as BaseConnection
from sqlalchemy import CursorResult
from sqlalchemy import Executable
from sqlalchemy import ScalarResult
from sqlalchemy import text
from sqlalchemy.engine.interfaces import CoreExecuteOptionsParameter
from sqlalchemy.engine.interfaces import _CoreAnyExecuteParams
from sqlalchemy.engine.interfaces import _CoreSingleExecuteParams
from sqlalchemy.sql.selectable import TypedReturnsRows


_T = TypeVar("_T", bound=Any)


class Connection(BaseConnection):
    @overload
    def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> _T | None: ...

    @overload
    def scalar(
        self,
        statement: Executable,
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> Any: ...

    @overload
    def scalar(
        self,
        statement: str,
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> Any: ...

    def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]] | Executable | str,
        parameters: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> Any:
        if isinstance(statement, str):
            statement = text(statement)

        return super().scalar(
            statement, parameters, execution_options=execution_options
        )

    @overload
    def scalars(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[_T]: ...

    @overload
    def scalars(
        self,
        statement: Executable,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[Any]: ...

    @overload
    def scalars(
        self,
        statement: str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[Any]: ...

    def scalars(
        self,
        statement: TypedReturnsRows[tuple[_T]] | Executable | str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> ScalarResult[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return super().scalars(
            statement, parameters, execution_options=execution_options
        )

    @overload
    def execute(
        self,
        statement: TypedReturnsRows[_T],
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[_T]: ...

    @overload
    def execute(
        self,
        statement: Executable,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[Any]: ...

    @overload
    def execute(
        self,
        statement: str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[Any]: ...

    def execute(
        self,
        statement: TypedReturnsRows[_T] | Executable | str,
        parameters: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: CoreExecuteOptionsParameter | None = None,
    ) -> CursorResult[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return super().execute(
            statement, parameters, execution_options=execution_options
        )

    def __enter__(self) -> Self:
        return self



from io import StringIO
from typing import Self

from cleo.io.inputs.string_input import StringInput
from cleo.io.outputs.buffered_output import BufferedOutput

from expanse.contracts.debug.exception_handler import ExceptionHandler
from expanse.core.application import Application
from expanse.core.console.gateway import Gateway


class TestingCommand:
    def __init__(
        self, app: Application, command: str, parameters: str | None = None
    ) -> None:
        self._app = app
        self._command = command
        self._output: BufferedOutput = BufferedOutput()
        self._return_code: int | None = None
        self._user_input: str | None = None

    @property
    def output(self) -> BufferedOutput:
        return self._output

    def with_user_input(self, user_input: str) -> Self:
        self._user_input = user_input

        return self

    def run(self, parameters: str | None = None) -> int:
        self._output.clear()

        full_command = self._command
        if parameters:
            full_command += " " + parameters

        handler = self._app.container.get(ExceptionHandler)
        input = StringInput(full_command)
        input.set_stream(StringIO())
        if self._user_input:
            input.stream.truncate(0)
            input.stream.seek(0)
            input.stream.write(self._user_input)
            input.stream.seek(0)
            input.interactive()

        with handler.raise_unhandled_exceptions():
            self._return_code = self._app.container.get(Gateway).handle(
                input, self._output
            )

        return self._return_code


class CommandTester:
    def __init__(self, app: Application) -> None:
        self._app = app

    def command(self, command: str, parameters: str | None = None) -> TestingCommand:
        return TestingCommand(self._app, command, parameters)



from __future__ import annotations

from typing import Generic
from typing import TypeVar

from expanse.common.routing.route import Route


RouteT = TypeVar("RouteT", bound=Route)


class RouteGroup(Generic[RouteT]):
    def __init__(
        self,
        name: str | None = None,
        prefix: str | None = None,
    ) -> None:
        self.name: str | None = name
        self.prefix: str | None = prefix
        self.routes: list[RouteT] = []

    def add_route(self, route: RouteT) -> None:
        self.routes.append(self._build_route(route))

    def _build_route(self, route: RouteT) -> RouteT:
        route_name = route.name
        if route_name is not None and self.name is not None:
            route_name = f"{self.name}.{route.name}"

        if self.prefix:
            if not route.path:
                route_path = self.prefix
            else:
                route_path = "/".join([self.prefix, route.path]).replace("//", "/")
        else:
            route_path = route.path

        return route.__class__(
            route_path, route.endpoint, methods=route.methods, name=route_name
        )



from collections.abc import Callable
from typing import Any
from typing import Literal
from typing import TypeVar

from sqlalchemy.orm import MappedColumn
from sqlalchemy.orm import mapped_column
from sqlalchemy.sql._typing import _AutoIncrementType
from sqlalchemy.sql._typing import _InfoType
from sqlalchemy.sql._typing import _TypeEngineArgument
from sqlalchemy.sql.base import SchemaEventTarget
from sqlalchemy.sql.base import _NoArg
from sqlalchemy.sql.schema import FetchedValue
from sqlalchemy.sql.schema import SchemaConst
from sqlalchemy.sql.schema import _ServerDefaultArgument


_T = TypeVar("_T")


def column(
    __name_pos: str | _TypeEngineArgument | SchemaEventTarget | None = None,
    __type_pos: _TypeEngineArgument | SchemaEventTarget | None = None,
    *args: SchemaEventTarget,
    init: bool = False,
    repr: _NoArg | bool = _NoArg.NO_ARG,
    default: Any | None = _NoArg.NO_ARG,
    default_factory: _NoArg | Callable[[], _T] = _NoArg.NO_ARG,
    compare: _NoArg | bool = _NoArg.NO_ARG,
    kw_only: _NoArg | bool = _NoArg.NO_ARG,
    nullable: bool
    | Literal[SchemaConst.NULL_UNSPECIFIED]
    | None = SchemaConst.NULL_UNSPECIFIED,
    primary_key: bool | None = False,
    deferred: _NoArg | bool = _NoArg.NO_ARG,
    deferred_group: str | None = None,
    deferred_raiseload: bool | None = None,
    use_existing_column: bool = False,
    name: str | None = None,
    type_: _TypeEngineArgument | None = None,
    autoincrement: _AutoIncrementType = "auto",
    doc: str | None = None,
    key: str | None = None,
    index: bool | None = None,
    unique: bool | None = None,
    info: _InfoType | None = None,
    onupdate: Any | None = None,
    insert_default: Any | None = _NoArg.NO_ARG,
    server_default: _ServerDefaultArgument | None = None,
    server_onupdate: FetchedValue | None = None,
    active_history: bool = False,
    quote: bool | None = None,
    system: bool = False,
    comment: str | None = None,
    sort_order: _NoArg | int = _NoArg.NO_ARG,
    **kw: Any,
) -> MappedColumn:
    return mapped_column(
        __name_pos,
        __type_pos,
        *args,
        init=init,
        repr=repr,
        default=default,
        default_factory=default_factory,
        compare=compare,
        kw_only=kw_only,
        nullable=nullable,
        primary_key=primary_key,
        deferred=deferred,
        deferred_group=deferred_group,
        deferred_raiseload=deferred_raiseload,
        use_existing_column=use_existing_column,
        name=name,
        type_=type_,
        autoincrement=autoincrement,
        doc=doc,
        key=key,
        index=index,
        unique=unique,
        info=info,
        onupdate=onupdate,
        insert_default=insert_default,
        server_default=server_default,
        server_onupdate=server_onupdate,
        active_history=active_history,
        quote=quote,
        system=system,
        comment=comment,
        sort_order=sort_order,
        **kw,
    )



from pathlib import Path

import pytest

from expanse.asynchronous.core.application import Application
from expanse.asynchronous.core.helpers import _use_container
from expanse.asynchronous.http.request import Request
from expanse.asynchronous.static.static import Static
from expanse.common.core.http.exceptions import HTTPException


FIXTURES_DIR = Path(__file__).parent / "fixtures"


async def test_get_returns_file_content_if_file_exists(app: Application) -> None:
    static = Static([FIXTURES_DIR], prefix="/static")

    async with app.container.create_scoped_container() as container:
        container.instance(Request, Request.create("http://example.com"))

        async with _use_container(container):
            response = await static.get("foo.txt")

    assert response.status_code == 200


async def test_get_returns_404_not_found_if_file_does_not_exist(
    app: Application,
) -> None:
    static = Static([FIXTURES_DIR], prefix="/static")

    async with app.container.create_scoped_container() as container:
        container.instance(Request, Request.create("http://example.com"))

        async with _use_container(container):
            with pytest.raises(HTTPException) as e:
                await static.get("bar.txt")

            assert e.value.status_code == 404



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import MutableMapping
from types import TracebackType
from typing import Any
from typing import Protocol


Environ = MutableMapping[str, Any]
ExcInfo = tuple[type[BaseException], BaseException, TracebackType | None]


class StartResponse(Protocol):
    def __call__(
        self,
        status: str,
        response_headers: tuple[str, str],
        exc_info: ExcInfo | None = None,
    ) -> None: ...


WSGIApp = Callable[[Environ, StartResponse], Iterable[bytes]]



from __future__ import annotations

import inspect
import re

from enum import Enum
from typing import TYPE_CHECKING
from typing import Self

from baize.routing import Route as BaseRoute


if TYPE_CHECKING:
    from expanse.types.routing import Endpoint


class Match(Enum):
    NONE = 0
    PARTIAL = 1
    FULL = 2


class Route(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: Endpoint,
        *,
        methods: list[str] | None = None,
        name: str | None = None,
    ) -> None:
        self.path: str = path
        self.endpoint: Endpoint = endpoint
        self.methods = methods or ["GET", "HEAD"]
        self.name: str | None = name
        self._param_names: set[str] | None = None
        self._signature: inspect.Signature | None = None

    @classmethod
    def get(cls, path: str, endpoint: Endpoint, *, name: str | None = None) -> Self:
        return cls(path, endpoint, methods=["GET", "HEAD"], name=name)

    @classmethod
    def post(cls, path: str, endpoint: Endpoint, *, name: str | None = None) -> Self:
        return cls(path, endpoint, methods=["POST"], name=name)

    @classmethod
    def put(cls, path: str, endpoint: Endpoint, *, name: str | None = None) -> Self:
        return cls(path, endpoint, methods=["PUT"], name=name)

    @classmethod
    def patch(cls, path: str, endpoint: Endpoint, *, name: str | None = None) -> Self:
        return cls(path, endpoint, methods=["PATCH"], name=name)

    @classmethod
    def delete(cls, path: str, endpoint: Endpoint, *, name: str | None = None) -> Self:
        return cls(path, endpoint, methods=["DELETE"], name=name)

    @classmethod
    def options(cls, path: str, endpoint: Endpoint, *, name: str | None = None) -> Self:
        return cls(path, endpoint, methods=["OPTIONS"], name=name)

    @classmethod
    def head(cls, path: str, endpoint: Endpoint, *, name: str | None = None) -> Self:
        return cls(path, endpoint, methods=["HEAD"], name=name)

    @property
    def param_names(self) -> set[str]:
        if self._param_names is None:
            self._param_names = set(re.findall(r"{([^:]*)(?::.*)?}", self.path))

        return self._param_names

    @property
    def signature(self) -> inspect.Signature:
        if self._signature is None:
            self._signature = inspect.signature(self.endpoint)

        return self._signature

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.path}", {self.endpoint})'



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Self

from expanse.common.routing.route_group import RouteGroup as BaseRouteGroup
from expanse.routing.route import Route


if TYPE_CHECKING:
    from expanse.core.http.middleware.middleware import Middleware
    from expanse.types.routing import Endpoint


class RouteGroup(BaseRouteGroup):
    def __init__(
        self,
        name: str | None = None,
        prefix: str | None = None,
    ) -> None:
        super().__init__(name, prefix)

        self._middlewares: list[type[Middleware]] = []

    def add_routes(self, routes: list[Route]) -> None:
        for route in routes:
            self.add_route(route)

    def get(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.get(path, endpoint, name=name)
        self.add_route(route)

        return route

    def post(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.post(path, endpoint, name=name)
        self.add_route(route)

        return route

    def put(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.patch(path, endpoint, name=name)
        self.add_route(route)

        return route

    def patch(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.patch(path, endpoint, name=name)
        self.add_route(route)

        return route

    def delete(
        self, path: str, endpoint: Endpoint, *, name: str | None = None
    ) -> Route:
        route = Route.delete(path, endpoint, name=name)
        self.add_route(route)

        return route

    def head(self, path: str, endpoint: Endpoint, *, name: str | None = None) -> Route:
        route = Route.head(path, endpoint, name=name)
        self.add_route(route)

        return route

    def options(
        self, path: str, endpoint: Endpoint, *, name: str | None = None
    ) -> Route:
        route = Route.options(path, endpoint, name=name)
        self.add_route(route)

        return route

    def middleware(self, *middlewares: type[Middleware]) -> Self:
        self._middlewares.extend(middlewares)

        return self

    def prepend_middleware(self, *middlewares: type[Middleware]) -> Self:
        self._middlewares = list(middlewares) + self._middlewares

        return self

    def _build_route(self, route: Route) -> Route:
        route = super()._build_route(route)

        if self._middlewares:
            route.prepend_middleware(*self._middlewares)

        return route

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        return



from collections.abc import MutableMapping
from http import HTTPStatus
from typing import Any


class HTTPException(Exception):
    def __init__(
        self,
        status_code: int,
        detail: str | None = None,
        headers: MutableMapping[str, Any] | None = None,
    ) -> None:
        if detail is None:
            detail = HTTPStatus(status_code).phrase

        self.status_code = status_code
        self.detail = detail
        self.headers = headers

    def __str__(self) -> str:
        return self.detail

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from pydantic_settings import BaseSettings

from expanse.common.configuration.config import Config
from expanse.common.support._utils import module_from_path


if TYPE_CHECKING:
    from expanse.asynchronous.core.application import Application


class LoadConfiguration:
    @classmethod
    async def bootstrap(cls, app: Application) -> None:
        config = Config({})

        base_files = await cls._get_base_configuration_files(app)
        for identifier, filepath in base_files.items():
            config[identifier] = await cls._load_configuration_file(filepath)

        files = await cls._get_configuration_files(app)

        for identifier, filepath in files.items():
            for name, value in (await cls._load_configuration_file(filepath)).items():
                config[f"{identifier}.{name}"] = value

        app.set_config(config)

    @classmethod
    async def _get_configuration_files(cls, app: Application) -> dict[str, Path]:
        files = {}

        for filepath in app.config_path.rglob("*.py"):
            files[
                filepath.relative_to(app.config_path)
                .with_suffix("")
                .as_posix()
                .replace("/", ":")
            ] = filepath

        return dict(sorted(files.items()))

    @classmethod
    async def _load_configuration_file(cls, path: Path) -> dict[str, Any]:
        module = module_from_path(path)

        if module is None:
            return {}

        if hasattr(module, "Config") and issubclass(module.Config, BaseSettings):
            config = module.Config().model_dump()
        else:
            config = module.config

        assert isinstance(config, dict)

        return config

    @classmethod
    async def _get_base_configuration_files(cls, app: Application) -> dict[str, Path]:
        files = {}

        config_assets_path = Path(__file__).parent.parent.parent.joinpath(
            "assets/config"
        )

        for filepath in config_assets_path.rglob("*.py"):
            files[
                filepath.relative_to(config_assets_path)
                .with_suffix("")
                .as_posix()
                .replace("/", ":")
            ] = filepath

        return dict(sorted(files.items()))



from __future__ import annotations

from expanse.asynchronous.http.response_adapter import ResponseAdapter
from expanse.asynchronous.support.service_provider import ServiceProvider


class HTTPServiceProvider(ServiceProvider):
    async def register(self) -> None:
        self._container.scoped(ResponseAdapter)



from typing import Annotated

import pytest

from expanse.database.session import Session
from expanse.http.helpers import json
from expanse.http.response import Response
from expanse.routing.router import Router
from expanse.testing.client import TestClient


def default_session(session: Session) -> Response:
    return json(
        (
            session.execute("SELECT id FROM my_table WHERE id = 'sqlite' LIMIT 1")
        ).scalar()
    )


def test_the_default_session_can_be_injected(
    router: Router, client: TestClient
) -> None:
    router.get("/default", default_session)

    response = client.get("/default")

    assert response.json() == "sqlite"


@pytest.mark.parametrize(
    "name",
    [
        "sqlite",
        "sqlite2",
        "postgresql",
        "postgresql_psycopg",
        "postgresql_psycopg2",
        "postgresql_pg8000",
    ],
)
def test_a_named_session_can_be_injected(
    router: Router, client: TestClient, name: str
) -> None:
    def named_session(session: Annotated[Session, name]) -> Response:
        return json(
            session.scalar(
                "SELECT id FROM my_table WHERE id = :id LIMIT 1", {"id": name}
            )
        )

    router.get("/named", named_session)

    response = client.get("/named")

    assert response.json() == name



from expanse.asynchronous.core.http.middleware.middleware import Middleware
from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.asynchronous.types.http.middleware import RequestHandler


class SimpleMiddleware(Middleware):
    async def handle(self, request: Request, next_call: RequestHandler) -> Response:
        response = await next_call(request)

        response.headers["X-Foo"] = "bar"

        return response


def test_route_group_with_middlewares(client: TestClient, router: Router) -> None:
    with router.group("api") as group:
        group.middleware(SimpleMiddleware)
        group.get("/", lambda: Response("Hello world!"))

    response = client.get("/")

    assert response.headers["X-Foo"] == "bar"



import pytest

from expanse.core.http.middleware.middleware import Middleware
from expanse.http.request import Request
from expanse.http.response import Response
from expanse.routing.router import Router
from expanse.testing.client import TestClient
from expanse.types.http.middleware import RequestHandler


class SimpleMiddleware(Middleware):
    def handle(self, request: Request, next_call: RequestHandler) -> Response:
        response = next_call(request)

        response.headers["X-Foo"] = "bar"

        return response


@pytest.mark.parametrize("method", ["get", "post", "put", "patch", "delete"])
def test_simple_routes_are_properly_registered(
    router: Router, client: TestClient, method: str
) -> None:
    getattr(router, method)("/", lambda: Response("Hello world!"))

    response = getattr(client, method)("/")

    assert response.text == "Hello world!"


@pytest.mark.parametrize("method", ["get", "post", "put", "patch", "delete"])
def test_routes_with_middleware_are_properly_registered(
    router: Router, client: TestClient, method: str
) -> None:
    getattr(router, method)("/", lambda: Response("Hello world!")).middleware(
        SimpleMiddleware
    )

    response = getattr(client, method)("/")

    assert response.text == "Hello world!"
    assert response.headers["X-Foo"] == "bar"



from pathlib import Path

import pytest

from expanse.console.commands.command import Command
from expanse.core.application import Application
from expanse.core.console.gateway import Gateway
from expanse.testing.command_tester import CommandTester


class FooCommand(Command):
    name = "foo"

    def handle(self) -> int:
        self.info("Foo")

        return 0


@pytest.fixture(autouse=True)
def setup_kernel(app: Application) -> Gateway:
    kernel = app.container.get(Gateway)
    kernel.add_command(FooCommand)
    kernel.add_command_paths([Path(__file__).parent.joinpath("fixtures/commands")])
    kernel.bootstrap()

    return kernel


def test_external_command_can_be_called(command_tester: CommandTester, app) -> None:
    command = command_tester.command("foo")

    assert command.run() == 0
    assert command.output.fetch().strip() == "Foo"


def test_commands_can_be_auto_discovered(command_tester: CommandTester) -> None:
    command = command_tester.command("foo bar")

    assert command.run() == 0
    assert command.output.fetch().strip() == "Foo Bar"



import re

from expanse.asynchronous.container.container import Container
from expanse.asynchronous.core.http.middleware.middleware import Middleware
from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.types.http.middleware import RequestHandler
from expanse.common.configuration.config import Config


class Cors:
    def __init__(
        self,
        allowed_origins: list[str] | None = None,
        allowed_origins_patterns: list[str] | None = None,
        allowed_methods: list[str] | None = None,
        allowed_headers: list[str] | None = None,
        exposed_headers: list[str] | None = None,
        supports_credentials: bool = False,
        max_age: int | None = 0,
    ) -> None:
        self._allowed_origins: list[str] = allowed_origins or []
        self._allowed_origins_patterns: list[str] = allowed_origins_patterns or []
        self._allowed_methods: list[str] = allowed_methods or []
        self._allowed_headers: list[str] = allowed_headers or []
        self._exposed_headers: list[str] = exposed_headers or []
        self._supports_credentials: bool = supports_credentials
        self._max_age: int | None = max_age

        self._allow_all_origins: bool = False
        self._allow_all_headers: bool = False
        self._allow_all_methods: bool = False

        self._normalize()

    def set_options(
        self,
        allowed_origins: list[str] | None = None,
        allowed_origins_patterns: list[str] | None = None,
        allowed_methods: list[str] | None = None,
        allowed_headers: list[str] | None = None,
        exposed_headers: list[str] | None = None,
        supports_credentials: bool = False,
        max_age: int | None = 0,
        **kwargs,
    ) -> None:
        self._allowed_origins = allowed_origins or []
        self._allowed_origins_patterns = allowed_origins_patterns or []
        self._allowed_methods = allowed_methods or []
        self._allowed_headers = allowed_headers or []
        self._exposed_headers = exposed_headers or []
        self._supports_credentials = supports_credentials
        self._max_age = max_age

        self._normalize()

    def is_cors_request(self, request: Request) -> bool:
        return "Origin" in request.headers

    def is_preflight_request(self, request: Request) -> bool:
        return (
            request.method == "OPTIONS"
            and "Access-Control-Request-Method" in request.headers
        )

    def handle_preflight_request(self, request: Request) -> Response:
        response = Response(status_code=204)

        return self.add_preflight_request_headers(response, request)

    def add_preflight_request_headers(
        self, response: Response, request: Request
    ) -> Response:
        self._configure_allowed_origins(response, request)

        if "Access-Control-Allow-Origin" in response.headers:
            self._configure_allow_credentials(response, request)

            self._configure_allowed_methods(response, request)

            self._configure_allowed_headers(response, request)

            self._configure_max_age(response, request)

        return response

    def add_actual_request_headers(
        self, response: Response, request: Request
    ) -> Response:
        self._configure_allowed_origins(response, request)

        if "Access-Control-Allow-Origin" in response.headers:
            self._configure_allow_credentials(response, request)

            self._configure_exposed_headers(response, request)

        return response

    def is_origin_allowed(self, request: Request) -> bool:
        if self._allow_all_origins:
            return True

        origin = request.headers.get("Origin", "")

        if origin in self._allowed_origins:
            return True

        for pattern in self._allowed_origins_patterns:
            if re.match(pattern, origin):
                return True

        return False

    def vary_header(self, response: Response, header: str) -> Response:
        if "Vary" not in response.headers:
            response.headers["Vary"] = header
        elif header not in response.headers["Vary"].split(", "):
            response.headers["Vary"] += f", {header}"

        return response

    def _normalize(self) -> None:
        # We normalize case first
        self._allowed_headers = [header.lower() for header in self._allowed_headers]
        self._allowed_methods = [method.lower() for method in self._allowed_methods]

        # Normalize ["*"] to true
        self._allow_all_origins = "*" in self._allowed_origins
        self._allow_all_headers = "*" in self._allowed_headers
        self._allow_all_methods = "*" in self._allowed_methods

        if not self._allow_all_origins:
            for origin in self._allowed_origins:
                if "*" in origin:
                    self._allowed_origins_patterns.append(
                        self._convert_wildcard_to_pattern(origin)
                    )

    def _convert_wildcard_to_pattern(self, pattern: str) -> str:
        pattern = re.escape(pattern)
        pattern = pattern.replace(r"\*", ".*")

        return f"^{pattern}"

    def _configure_allowed_origins(self, response: Response, request: Request) -> None:
        if self._allow_all_origins and not self._supports_credentials:
            # Safe+cacheable, allow everything
            response.headers["Access-Control-Allow-Origin"] = "*"
        elif self._is_single_origin_allowed():
            # Single origins can be safely set
            response.headers["Access-Control-Allow-Origin"] = self._allowed_origins[0]
        else:
            if self.is_cors_request(request) and self.is_origin_allowed(request):
                response.headers["Access-Control-Allow-Origin"] = request.headers[
                    "Origin"
                ]

            self.vary_header(response, "Origin")

    def _is_single_origin_allowed(self) -> bool:
        if self._allow_all_origins or len(self._allowed_origins_patterns) > 0:
            return False

        return len(self._allowed_origins) == 1

    def _configure_allow_credentials(
        self, response: Response, request: Request
    ) -> None:
        if self._supports_credentials:
            response.headers["Access-Control-Allow-Credentials"] = "true"

    def _configure_allowed_methods(self, response: Response, request: Request) -> None:
        if self._allow_all_methods:
            allow_methods = request.headers.get(
                "Access-Control-Request-Method", ""
            ).upper()
            self.vary_header(response, "Access-Control-Request-Method")
        else:
            allow_methods = ", ".join(self._allowed_methods).upper()

        response.headers["Access-Control-Allow-Methods"] = allow_methods

    def _configure_allowed_headers(self, response: Response, request: Request) -> None:
        if self._allow_all_headers:
            allow_headers = request.headers.get("Access-Control-Request-Headers", "")
            self.vary_header(response, "Access-Control-Request-Headers")
        else:
            allow_headers = ", ".join(self._allowed_headers)

        response.headers["Access-Control-Allow-Headers"] = allow_headers

    def _configure_max_age(self, response: Response, request: Request) -> None:
        if self._max_age is not None:
            response.headers["Access-Control-Max-Age"] = str(self._max_age)

    def _configure_exposed_headers(self, response: Response, request: Request) -> None:
        if self._exposed_headers:
            response.headers["Access-Control-Expose-Headers"] = ", ".join(
                self._exposed_headers
            )


class ManageCors(Middleware):
    def __init__(self, container: Container, config: Config) -> None:
        self._container: Container = container
        self._cors: Cors = Cors()
        self._config: Config = config

    async def handle(self, request: Request, next_call: RequestHandler) -> Response:
        """
        Handle the incoming request.
        """
        if not self._has_matching_path(request):
            return await next_call(request)

        self._cors.set_options(**(await self._container.get(Config)).get("cors", {}))

        if self._cors.is_preflight_request(request):
            response = self._cors.handle_preflight_request(request)

            self._cors.vary_header(response, "Access-Control-Request-Method")

            return response

        response = await next_call(request)

        if request.method == "OPTIONS":
            self._cors.vary_header(response, "Access-Control-Request-Method")

        self._cors.add_actual_request_headers(response, request)

        return response

    def _has_matching_path(self, request: Request) -> bool:
        paths: list[str] = self._config.get("cors", {}).get("paths", [])

        for path in paths:
            if path != "/":
                path = path.lstrip("/")

            if request.url.is_(path) or request.url.path_is(path):
                return True

        return False


__all__ = ["ManageCors"]



from typing import Any

import pytest

from expanse.http.request import Request


@pytest.mark.parametrize(
    "content_type,expected",
    [("application/json", True), ("application/foo+json", True), ("text/html", False)],
)
def test_is_json(environ: dict[str, Any], content_type: str, expected: bool) -> None:
    environ["CONTENT_TYPE"] = content_type

    request = Request(environ)

    assert request.is_json() is expected


@pytest.mark.parametrize(
    "header,expected",
    [
        ("application/json", True),
        ("application/json; charset=utf-8", True),
        ("application/foo+json", True),
        ("application/json, text/html", True),
        ("text/html", False),
        ("application/json;q=0.1, text/html;q=0.8", False),
        ("application/json;q=0.8, text/html;q=0.8", True),
    ],
)
def test_wants_json(environ: dict[str, Any], header: str, expected: bool) -> None:
    environ["HTTP_ACCEPT"] = header

    request = Request(environ)

    assert request.wants_json() is expected



from __future__ import annotations

import os
import re
import shutil
import sys

from abc import ABC
from abc import abstractmethod
from contextlib import suppress
from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar

from cleo.exceptions import CleoCommandNotFoundError
from cleo.exceptions import CleoError
from cleo.exceptions import CleoLogicError
from cleo.exceptions import CleoNamespaceNotFoundError
from cleo.exceptions import CleoUserError
from cleo.formatters.style import Style
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.argv_input import ArgvInput
from cleo.io.inputs.definition import Definition
from cleo.io.inputs.option import Option
from cleo.io.io import IO
from cleo.io.outputs.output import Verbosity
from cleo.io.outputs.stream_output import StreamOutput
from cleo.ui.ui import UI

from expanse.__version__ import __version__
from expanse.common.console.command import Command
from expanse.common.console.command_loader import CommandLoader


if TYPE_CHECKING:
    from cleo.io.inputs.input import Input
    from cleo.io.outputs.output import Output
    from crashtest.solution_providers.solution_provider_repository import (
        SolutionProviderRepository,
    )

TCommand = TypeVar("TCommand", bound=Command)


class Console(Generic[TCommand], ABC):
    def __init__(self) -> None:
        self._name = "Craft"
        self._version = __version__
        self._display_name: str | None = None
        self._terminal = shutil.get_terminal_size()
        self._default_command = "list"
        self._single_command = False
        self._want_helps = False
        self._definition: Definition | None = None
        self._catch_exceptions = True
        self._auto_exit = True
        self._initialized = False
        self._ui: UI | None = None

        self._command_loader: CommandLoader[TCommand] = CommandLoader({})
        self._solution_provider_repository: SolutionProviderRepository | None = None

        self._commands: dict[str, TCommand] = {}
        self._running_command: TCommand | None = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def display_name(self) -> str:
        if self._display_name is None:
            return re.sub(r"[\s\-_]+", " ", self._name).title()

        return self._display_name

    @property
    def version(self) -> str:
        return self._version

    @property
    def long_version(self) -> str:
        if self._name:
            if self._version:
                return f"<b>{self.display_name}</b> (version <c1>{self._version}</c1>)"

            return f"<b>{self.display_name}</b>"

        return "<b>Console</b> application"

    @property
    def definition(self) -> Definition:
        if self._definition is None:
            self._definition = self._default_definition

        if self._single_command:
            definition = self._definition
            definition.set_arguments([])

            return definition

        return self._definition

    @property
    def default_commands(self) -> list[TCommand]:
        return []

    @property
    def help(self) -> str:
        return self.long_version

    @property
    def ui(self) -> UI:
        if self._ui is None:
            self._ui = self._get_default_ui()

        return self._ui

    @property
    def command_loader(self) -> CommandLoader[TCommand]:
        return self._command_loader

    def set_name(self, name: str) -> None:
        self._name = name

    def set_display_name(self, display_name: str) -> None:
        self._display_name = display_name

    def set_version(self, version: str) -> None:
        self._version = version

    def set_ui(self, ui: UI) -> None:
        self._ui = ui

    def set_command_loader(self, command_loader: CommandLoader[TCommand]) -> None:
        self._command_loader = command_loader

    def auto_exits(self, auto_exits: bool = True) -> None:
        self._auto_exit = auto_exits

    def is_auto_exit_enabled(self) -> bool:
        return self._auto_exit

    def are_exceptions_caught(self) -> bool:
        return self._catch_exceptions

    def catch_exceptions(self, catch_exceptions: bool = True) -> None:
        self._catch_exceptions = catch_exceptions

    def is_single_command(self) -> bool:
        return self._single_command

    def set_solution_provider_repository(
        self, solution_provider_repository: SolutionProviderRepository
    ) -> None:
        self._solution_provider_repository = solution_provider_repository

    def add(self, command: TCommand) -> TCommand | None:
        self._init()

        command.set_console(self)

        if not command.enabled:
            command.set_console()

            return None

        if not command.name:
            raise CleoLogicError(
                f'The command "{command.__class__.__name__}" cannot have an empty name'
            )

        self._commands[command.name] = command

        for alias in command.aliases:
            self._commands[alias] = command

        return command

    def get(self, name: str) -> TCommand:
        self._init()

        if not self.has(name):
            raise CleoCommandNotFoundError(name)

        if name not in self._commands:
            # The command was registered in a different name in the command loader
            raise CleoCommandNotFoundError(name)

        command = self._commands[name]

        if self._want_helps:
            self._want_helps = False

            return self.get_help_command(command)

        return command

    @abstractmethod
    def get_help_command(self, command: TCommand) -> TCommand: ...

    def has(self, name: str) -> bool:
        self._init()

        if name in self._commands:
            return True

        if not self._command_loader:
            return False

        return bool(
            self._command_loader.has(name) and self.add(self._command_loader.get(name))
        )

    def get_namespaces(self) -> list[str]:
        namespaces = []
        seen = set()

        for command in self.all().values():
            if command.hidden or not command.name:
                continue

            for namespace in self._extract_all_namespaces(command.name):
                if namespace in seen:
                    continue

                namespaces.append(namespace)
                seen.add(namespace)

            for alias in command.aliases:
                for namespace in self._extract_all_namespaces(alias):
                    if namespace in seen:
                        continue

                    namespaces.append(namespace)
                    seen.add(namespace)

        return namespaces

    def find_namespace(self, namespace: str) -> str:
        all_namespaces = self.get_namespaces()

        if namespace not in all_namespaces:
            raise CleoNamespaceNotFoundError(namespace, all_namespaces)

        return namespace

    def find(self, name: str) -> TCommand:
        self._init()

        if self.has(name):
            return self.get(name)

        all_commands = []
        if self._command_loader:
            all_commands += self._command_loader.names

        all_commands += [
            name for name, command in self._commands.items() if not command.hidden
        ]

        raise CleoCommandNotFoundError(name, all_commands)

    def all(self, namespace: str | None = None) -> dict[str, TCommand]:
        self._init()

        if namespace is None:
            commands = self._commands.copy()
            if not self._command_loader:
                return commands

            for name in self._command_loader.names:
                if name not in commands and self.has(name):
                    commands[name] = self.get(name)

            return commands

        commands = {}

        for name, command in self._commands.items():
            if namespace == self.extract_namespace(name, name.count(" ") + 1):
                commands[name] = command

        if self._command_loader:
            for name in self._command_loader.names:
                if (
                    name not in commands
                    and namespace == self.extract_namespace(name, name.count(" ") + 1)
                    and self.has(name)
                ):
                    commands[name] = self.get(name)

        return commands

    def _setup_command(self, io: IO) -> None:
        definition = self.definition
        input_definition = Definition()
        for argument in definition.arguments:
            if argument.name == "command":
                argument = Argument(
                    "command",
                    required=True,
                    is_list=True,
                    description=definition.argument("command").description,
                )

            input_definition.add_argument(argument)

        input_definition.set_options(definition.options)

        # Errors must be ignored, full binding/validation
        # happens later when the command is known.
        with suppress(CleoError):
            # Makes ArgvInput.first_argument() able to
            # distinguish an option from an argument.
            io.input.bind(input_definition)

        name = self._get_command_name(io)
        if io.input.has_parameter_option(["--help", "-h"], True):
            if not name:
                name = "help"
                io.set_input(ArgvInput(["console", "help", self._default_command]))
            else:
                self._want_helps = True

        if not name:
            name = self._default_command
            definition = self.definition
            arguments = definition.arguments
            if not definition.has_argument("command"):
                arguments.append(
                    Argument(
                        "command",
                        required=False,
                        description=definition.argument("command").description,
                        default=name,
                    )
                )
            definition.set_arguments(arguments)

        self._running_command = None
        command = self.find(name)

        self._running_command = command

        if " " in name and isinstance(io.input, ArgvInput):
            # If the command is namespaced we rearrange
            # the input to parse it as a single argument
            argv = io.input._tokens[:]

            start: int = 0
            if io.input.script_name is not None:
                argv.insert(0, io.input.script_name)
                start = 1

            namespace = name.split(" ")[:-1]
            index = None
            for i, arg in enumerate(argv):
                if i < start:
                    continue

                if namespace and arg == namespace[0] and len(namespace) == 1:
                    argv[i] = name
                    index = i if index is None else index
                    break
                elif namespace and arg == namespace[0]:
                    namespace.pop(0)
                    index = i

            if index is not None:
                del argv[index + 1 : index + name.count(" ") + start]

            stream = io.input.stream
            interactive = io.input.is_interactive()
            io.set_input(ArgvInput(argv))
            io.input.set_stream(stream)
            io.input.interactive(interactive)

    def create_io(
        self,
        input: Input | None = None,
        output: Output | None = None,
        error_output: Output | None = None,
    ) -> IO:
        if input is None:
            input = ArgvInput()
            input.set_stream(sys.stdin)

        if output is None:
            output = StreamOutput(sys.stdout)

        if error_output is None:
            error_output = StreamOutput(sys.stderr)

        io = IO(input, output, error_output)

        # Set our own CLI styles
        formatter = io.output.formatter
        formatter.set_style("c1", Style("cyan"))
        formatter.set_style("c2", Style("default", options=["bold"]))
        formatter.set_style("info", Style("blue"))
        formatter.set_style("comment", Style("green"))
        formatter.set_style("warning", Style("yellow"))
        formatter.set_style("debug", Style("default", options=["dark"]))
        formatter.set_style("success", Style("green"))

        return io

    def render_error(self, error: Exception, io: IO) -> None:
        from cleo.ui.exception_trace import ExceptionTrace

        trace = ExceptionTrace(
            error, solution_provider_repository=self._solution_provider_repository
        )
        simple = not io.is_verbose() or isinstance(error, CleoUserError)
        trace.render(io.error_output, simple)

    def _configure_io(self, io: IO) -> None:
        if io.input.has_parameter_option("--ansi", True):
            io.decorated(True)
        elif io.input.has_parameter_option("--no-ansi", True):
            io.decorated(False)

        if io.input.has_parameter_option(["--no-interaction", "-n"], True) or (
            io.input._interactive is None
            and io.input.stream
            and not io.input.stream.isatty()
        ):
            io.interactive(False)

        shell_verbosity = int(os.getenv("SHELL_VERBOSITY", 0))
        if shell_verbosity == -1:
            io.set_verbosity(Verbosity.QUIET)
        elif shell_verbosity == 1:
            io.set_verbosity(Verbosity.VERBOSE)
        elif shell_verbosity == 2:
            io.set_verbosity(Verbosity.VERY_VERBOSE)
        elif shell_verbosity == 3:
            io.set_verbosity(Verbosity.DEBUG)
        else:
            shell_verbosity = 0

        if io.input.has_parameter_option(["--quiet", "-q"], True):
            io.set_verbosity(Verbosity.QUIET)
            shell_verbosity = -1
        else:
            if io.input.has_parameter_option("-vvv", True):
                io.set_verbosity(Verbosity.DEBUG)
                shell_verbosity = 3
            elif io.input.has_parameter_option("-vv", True):
                io.set_verbosity(Verbosity.VERY_VERBOSE)
                shell_verbosity = 2
            elif io.input.has_parameter_option(
                "-v", True
            ) or io.input.has_parameter_option("--verbose", only_params=True):
                io.set_verbosity(Verbosity.VERBOSE)
                shell_verbosity = 1

        if shell_verbosity == -1:
            io.interactive(False)

    @property
    def _default_definition(self) -> Definition:
        return Definition(
            [
                Argument(
                    "command",
                    required=True,
                    description="The command to execute.",
                ),
                Option(
                    "--help",
                    "-h",
                    flag=True,
                    description=(
                        "Display help for the given command. "
                        "When no command is given display help for "
                        f"the <info>{self._default_command}</info> command."
                    ),
                ),
                Option(
                    "--quiet", "-q", flag=True, description="Do not output any message."
                ),
                Option(
                    "--verbose",
                    "-v|vv|vvv",
                    flag=True,
                    description=(
                        "Increase the verbosity of messages: "
                        "1 for normal output, 2 for more verbose "
                        "output and 3 for debug."
                    ),
                ),
                Option(
                    "--version",
                    "-V",
                    flag=True,
                    description="Display this application version.",
                ),
                Option("--ansi", flag=True, description="Force ANSI output."),
                Option("--no-ansi", flag=True, description="Disable ANSI output."),
                Option(
                    "--no-interaction",
                    "-n",
                    flag=True,
                    description="Do not ask any interactive question.",
                ),
            ]
        )

    def _get_command_name(self, io: IO) -> str | None:
        if self._single_command:
            return self._default_command

        if "command" in io.input.arguments and io.input.argument("command"):
            candidates: list[str] = []
            for command_part in io.input.argument("command"):
                if candidates:
                    candidates.append(candidates[-1] + " " + command_part)
                else:
                    candidates.append(command_part)

            for candidate in reversed(candidates):
                if self.has(candidate):
                    return candidate

        return io.input.first_argument

    def extract_namespace(self, name: str, limit: int | None = None) -> str:
        parts = name.split(" ")[:-1]
        return " ".join(parts[:limit])

    def _get_default_ui(self) -> UI:
        from cleo.ui.progress_bar import ProgressBar

        io = self.create_io()
        return UI([ProgressBar(io)])

    def _extract_all_namespaces(self, name: str) -> list[str]:
        parts = name.split(" ")[:-1]
        namespaces: list[str] = []

        for part in parts:
            namespaces.append(namespaces[-1] + " " + part if namespaces else part)

        return namespaces

    def _init(self) -> None:
        if self._initialized:
            return

        self._initialized = True

        for command in self.default_commands:
            self.add(command)



from typing import TYPE_CHECKING

from expanse.asynchronous.console.commands.command import Command
from expanse.common.database.migration.utils import configure_alembic_loggers


if TYPE_CHECKING:
    from cleo.io.io import IO


class MigrationCommand(Command):
    async def run(self, io: "IO") -> int:
        configure_alembic_loggers(io)

        return await super().run(io)
